import inspect
import json
from typing import Any, Callable, Dict
from urllib.parse import urlparse, urlunparse

from plexapi.server import PlexServer
from typing_extensions import Literal, TypedDict
from websocket import WebSocketApp


class NotificationContainerDict(TypedDict):
    """Type of the underlying dictionary sent in each WebSocket frame.

    See _MessageDict.

    If type is 'playing', the dictionary also has an entry with key
    'PlaySessionStateNotification' of type PlaybackNotification.
    """

    type: str
    size: int

    # if type == 'playing':
    #     PlaySessionStateNotification: PlaybackNotification


class PlaybackNotification(TypedDict):
    """Type of NotificationContainerDict['PlaySessionStateNotification']."""

    sessionKey: str  # Not an int!
    guid: str  # Can be the empty string.
    ratingKey: str
    url: str  # Can be the empty string.
    key: str
    viewOffset: int  # In milliseconds.
    playQueueItemID: int
    state: Literal['buffering', 'playing', 'paused', 'stopped']


class _MessageDict(TypedDict):
    """The format of each WebSocket frame emitted by Plex once parsed."""
    NotificationContainer: NotificationContainerDict


class LoudWebSocketApp(WebSocketApp):
    def _callback(self, callback, *args):
        """The base implementation silences exceptions, unlike this one."""
        if callback:
            if inspect.ismethod(callback):
                callback(*args)
            else:
                callback(self, *args)


class NotificationListener:
    """Cleaner implementation of plexapi.alert.AlertListener.

    By default, it uses an implementation of websocket.WebSocketApp that doesn't
    silence exceptions. It also doesn't needlessly spawn a new thread.
    """

    def __init__(self, server: PlexServer, callback: Callable[[NotificationContainerDict], None]):
        self._server = server
        self._callback = callback

    def _get_ws_url(self) -> str:
        endpoint = '/:/websockets/notifications'
        http_parse = urlparse(self._server.url(endpoint, includeToken=True))
        return urlunparse(http_parse._replace(scheme='ws'))

    def run_forever(self):
        """Listens on the WebSocket and blocks indefinitely."""
        ws_url = self._get_ws_url()
        ws_app = LoudWebSocketApp(ws_url, on_message=self._on_message, on_error=self._on_error)
        ws_app.run_forever()

    def _on_message(self, message: str):
        msg_dict: _MessageDict = json.loads(message)
        self._callback(msg_dict['NotificationContainer'])

    def _on_error(self, e: Exception):
        raise e



from abc import ABC, abstractmethod
import logging
import threading
from typing import Dict, List, Tuple
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse
from uuid import UUID

from plexapi.client import DEFAULT_MTYPE, PlexClient
from plexapi.server import PlexServer
import pychromecast
from pychromecast.controllers.plex import PlexController
import requests
from wrapt.decorators import synchronized
from zeroconf import Zeroconf

from .sessions import Session


logger = logging.getLogger(__name__)


def _removesuffix(s: str, suffix: str) -> str:
    if s.endswith(suffix):
        return s[:-len(suffix)]
    return s


class Seekable(ABC):
    @abstractmethod
    def seek(self, offset_ms: int):
        pass


class SeekablePlexClient(Seekable):
    _TIMEOUT_SUFFIX = '-timeout'

    def __init__(self, client: PlexClient, timeout_sec: float = 5):
        self._client = client
        self._timeout_sec = timeout_sec

        # Save the original PlexClient.query() for future monkey patching in
        # seek(). Save it here and not in seek() to avoid a potential stack
        # overflow (with nested method calls) at run-time if we patch a lot.
        self._original_query = self._client.query

        # HACK: Monkey patch PlexClient.query() so that it takes the specified
        # timeout into account for the seekTo command.
        self._client.query = self._patched_query

    # Same signature as PlexClient.query().
    def _patched_query(self, path, method=None, headers=None, timeout=None, **kwargs):
        """Patched implementation of self._client.query()."""
        p = urlparse(path)
        new_qsl: List[Tuple[str, str]] = []

        for k, v in parse_qsl(p.query):
            if k == 'type':
                new_v = _removesuffix(v, self._TIMEOUT_SUFFIX)
                if new_v != v:
                    # The suffix was present, override the timeout.
                    timeout = self._timeout_sec
                new_qsl.append((k, new_v))
            else:
                new_qsl.append((k, v))

        new_query = urlencode(new_qsl, doseq=True)
        new_path = urlunparse(p._replace(query=new_query))

        return self._original_query(
            path=new_path,
            method=method,
            headers=headers,
            timeout=timeout,
            **kwargs
        )

    def seek(self, offset_ms: int):
        """Sends the seeking command in a non-blocking fashion.

        When tested against an iPhone client (iOS 14.3, Plex for iOS 7.11, Plex
        Media Server 1.21.1.3830), the seeking command takes a long time (over
        15 seconds) to issue a response, even though the client successfully
        seeks in less than a second. Therefore, we send the seeking command in
        a new thread and we only log what happens.
        """
        def _seek():
            def log_timeout_warning():
                # About "Advertize as player": If the user disables that setting
                # while Skippex is running, seeking will timeout (and not
                # happen), even though PlexSeekableProvider found the client.
                logger.warning(
                    f'Seeking command timed out for {self._client}, but '
                    f'seeking might still have happened. If not, please ensure '
                    f'that the "Advertize as player" setting is enabled for '
                    f'your client.'
                )

            try:
                # HACK: We add a suffix to mtype to signal to the patched method
                # to use the timeout set on this instance. This is the only way
                # we have to "pass a message" to PlexClient.query() from this
                # call.
                self._client.seekTo(offset_ms, mtype=DEFAULT_MTYPE+self._TIMEOUT_SUFFIX)
            except requests.Timeout:
                log_timeout_warning()
            except requests.ConnectionError as e:
                # See https://github.com/psf/requests/issues/5430.
                if 'timed out' in str(e):
                    log_timeout_warning()
                raise
            except Exception:
                logger.exception(f'Seeking failed for {self._client}')
            else:
                logger.debug(f'Seeking succeeded for {self._client}')

        thread = threading.Thread(target=_seek, daemon=True)
        thread.start()
        logger.debug(f'Sent seeking command to {self._client}')


class SeekableChromecastAdapter(Seekable):
    def __init__(self, plex_ctrl: PlexController):
        self._plex_ctrl = plex_ctrl

    def seek(self, offset_ms: int):
        self._plex_ctrl.seek(offset_ms / 1000)


class SeekableNotFoundError(Exception):
    def has_plex_player_not_found(self) -> bool:
        return isinstance(self, PlexPlayerNotFoundError)


class PlexPlayerNotFoundError(SeekableNotFoundError):
    pass


class SeekableNotFoundErrorChain(SeekableNotFoundError):
    def __init__(self, exceptions: List[SeekableNotFoundError], *args: object):
        super().__init__(exceptions, *args)
        self.exceptions = exceptions

    def has_plex_player_not_found(self) -> bool:
        for e in self.exceptions:
            if isinstance(e, PlexPlayerNotFoundError):
                return True
        return False


class SeekableProvider(ABC):
    @abstractmethod
    def provide_seekable(self, session: Session) -> Seekable:
        """Raises SeekableNotFoundError if no Seekable could be found."""
        pass


class SeekableProviderChain(SeekableProvider):
    def __init__(self, providers: List[SeekableProvider]):
        self._providers = providers

    def provide_seekable(self, session: Session) -> Seekable:
        exceptions = []
        for provider in self._providers:
            try:
                return provider.provide_seekable(session)
            except SeekableNotFoundError as e:
                exceptions.append(e)
        else:
            raise SeekableNotFoundErrorChain(exceptions)


class PlexSeekableProvider(SeekableProvider):
    def __init__(self, server: PlexServer):
        self._server = server

    def provide_seekable(self, session: Session) -> Seekable:
        sess_machine_id = session.player.machineIdentifier
        # NOTE: Have to "advertize as player" in order to be considered a client by Plex.
        client: PlexClient
        for client in self._server.clients():
            if client.machineIdentifier == sess_machine_id:
                return SeekablePlexClient(client)
        raise PlexPlayerNotFoundError(f'could not find Plex player with machine ID {sess_machine_id}')


class ChromecastNotFoundError(Exception):
    pass


class ChromecastMonitor:
    # The callbacks are called from a thread different from the main thread.

    def __init__(self, listener: pychromecast.CastListener, zconf: Zeroconf):
        self._listener = listener
        self._zconf = zconf
        self._chromecasts: Dict[UUID, pychromecast.Chromecast] = {}

    @synchronized
    def get_chromecast_by_ip(self, ip: str) -> pychromecast.Chromecast:
        for cc in self._chromecasts.values():
            if cc.socket_client.host == ip:
                return cc
        logging.debug(f'Discovered Chromecasts: {self._chromecasts}')
        raise ChromecastNotFoundError(f'could not find Chromecast with address {ip}')

    @synchronized
    def add_callback(self, uuid: UUID, name: str):
        service = self._listener.services[uuid]
        chromecast = pychromecast.get_chromecast_from_service(service, self._zconf)
        chromecast.wait()
        self._chromecasts[uuid] = chromecast
        logger.debug(f'Discovered new Chromecast: {chromecast}')

    @synchronized
    def update_callback(self, uuid: UUID, name: str):
        # No-op.
        pass

    @synchronized
    def remove_callback(self, uuid: UUID, name: str, service):
        chromecast = self._chromecasts.pop(uuid)
        logger.debug(f'Removed discovered Chromecast: {chromecast}')


class ChromecastSeekableProvider(SeekableProvider):
    def __init__(self, monitor: ChromecastMonitor):
        self._monitor = monitor

    def provide_seekable(self, session: Session) -> Seekable:
        try:
            chromecast = self._monitor.get_chromecast_by_ip(session.player.address)
        except ChromecastNotFoundError as e:
            raise SeekableNotFoundError(
                f'could not find Chromecast with address {session.player.address}'
            ) from e

        plex_ctrl = PlexController()
        chromecast.register_handler(plex_ctrl)
        return SeekableChromecastAdapter(plex_ctrl)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['skippex']

package_data = \
{'': ['*']}

install_requires = \
['PlexAPI>=4.2.0,<5.0.0',
 'PyChromecast>=7.7.1,<8.0.0',
 'pid>=3.0.4,<4.0.0',
 'requests>=2.25.1,<3.0.0',
 'typing-extensions>=3.7.4,<4.0.0',
 'websocket_client>=0.57.0,<0.58.0',
 'wrapt>=1.12.1,<2.0.0',
 'xdg>=5.0.1,<6.0.0',
 'zeroconf>=0.28.8,<0.29.0']

extras_require = \
{':python_version < "3.7"': ['dataclasses>=0.7,<0.8']}

entry_points = \
{'console_scripts': ['skippex = skippex.cmd:main']}

setup_kwargs = {
    'name': 'skippex',
    'version': '0.2.0',
    'description': "Automatic 'skip intro' for Plex with Chromecast support",
    'long_description': '# Skippex\n\nSkippex skips intros automatically for you on Plex, with support for the\nChromecast.\n\n**IMPORTANT NOTE**: This is still pretty much beta software. Expect bugs and\nplease report them!\n\n## Installation\n\nInstalling Skippex through **Docker** is the easiest way to get started:\n\n```console\n$ docker pull ghcr.io/sprt/skippex\n```\n\n*Docker-compose example coming soon.*\n\nIf you prefer not to use Docker, you can also use [**pipx**][pipx], which will\ninstall Skippex in its own virtual environment:\n\n```console\n$ pipx install skippex\n```\n\nOr you can just use **pip**:\n\n```console\n$ pip install --user skippex\n```\n\n[pipx]: https://pipxproject.github.io/pipx/\n\n## Usage\n\nThe first time you use Skippex, you\'ll first have to authorize the application\nwith Plex using the following command. This will open a new tab in your Web\nbrowser allowing you to authenticate and authorize the application to access\nyour Plex account.\n\n<table>\n  <tr>\n    <th>Docker</th>\n    <th>pipx & pip</th>\n  </tr>\n  <tr>\n    <td>\n      <code>$ docker run -v skippex:/config ghcr.io/sprt/skippex auth</code>\n    </td>\n    <td>\n      <code>$ skippex auth</code>\n    </td>\n  </tr>\n</table>\n\nOnce that\'s done, you can simply run Skippex and it\'ll start monitoring your\nplayback sessions and automatically skip intros for you on supported devices:\n\n<table>\n  <tr>\n    <th>Docker</th>\n    <th>pipx & pip</th>\n  </tr>\n  <tr>\n    <td>\n      <code>$ docker run -v skippex:/config ghcr.io/sprt/skippex run</code>\n    </td>\n    <td>\n      <code>$ skippex run</code>\n    </td>\n  </tr>\n</table>\n\nEt voilà! When this command says "Ready", Skippex is monitoring your shows and\nwill automatically skip intros for you.\n\n## Things to know\n\n * **Clients need to have "Advertize as player" enabled.**\n * Only skips once per playback session.\n * Only tested for one account on the local network.\n * Might only work on the local network for standard Plex clients.\n * Only works on the local network for Chromecasts.\n * Solely based on the intro markers detected by Plex; Skippex does not attempt\n   to detect intros itself.\n\n## Tested and supported players\n\n * Plex Web App\n * Plex for iOS (both iPhone and iPad)\n * Chromecast v3\n\nThe NVIDIA SHIELD might be supported as well, but I don\'t have one so I can\'t\ntest it. Other players might also be supported. In any case, please inform me\nby [creating a new issue][new_issue], so I can add your player to this list.\n\n[new_issue]: https://github.com/sprt/skippex/issues/new\n\n## Known issues\n\n * With a Chromecast, when seeking to a position, the WebSocket only receives\n   the notification 10 seconds later. Likewise, the HTTP API starts returning\n   the correct position only after 10 seconds. This means that if, before the\n   intro, the user seeks to within 10 seconds of the intro, they may view it for\n   a few seconds (before the notification comes in and saves us).\n\n   One workaround would be to listen to Chromecast status updates using\n   `pychromecast`, but that would necessitate a rearchitecture of the code.\n',
    'author': 'sprt',
    'author_email': 'hellosprt@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sprt/skippex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
